import os
os.environ['MONGOEASY_CONNECTION_STRING'] = 'mongodb://localhost:27017'
os.environ['MONGOEASY_DATABASE_NAME'] = 'mongeasy_pytest'
from pymongo import MongoClient
import pytest


from mongeasy import create_document_class, registry, disconnect, Query


@pytest.fixture(scope='function')
def clean_mongo():
    # Connect to MongoDB
    client = MongoClient('mongodb://localhost:27017')
    db = client['mongeasy_pytest']

    # Clean all collections in the test database
    for collection_name in db.list_collection_names():
        db[collection_name].delete_many({})

def test_connect_plugin(clean_mongo):
    
    from mongeasy import create_document_class, registry, disconnect, Query
    Connection = create_document_class('Connection', 'connection')
    class ConnectPlugin:
        def __init__(self):
            self.status = {
                'before_connect': False,
                'after_connect': False,
                'before_close': False,
                'after_close': False,
            }

        def before_connect(self):
            self.status['before_connect'] = True

        def after_connect(self):
            self.status['after_connect'] = True

        def before_close(self):
            self.status['before_close'] = True

        def after_close(self):
            self.status['after_close'] = True

    cp = ConnectPlugin()
    registry.register_plugin(cp)
    disconnect()
    assert cp.status['before_connect'] == False
    assert cp.status['after_connect'] == False
    assert cp.status['before_close'] == True
    assert cp.status['after_close'] == True


def test_init_document_plugin(clean_mongo):
        
        from mongeasy import create_document_class, registry, disconnect, Query
        User = create_document_class('User', 'users')
        class InitDocumentPlugin:
            def __init__(self):
                self.status = {
                    'before_init_document': False,
                    'after_init_document': False,
                }
    
            def before_init_document(self, *args, **kwargs):
                self.status['before_init_document'] = True
    
            def after_init_document(self, data):
                self.status['after_init_document'] = True
    
        idp = InitDocumentPlugin()
        registry.register_plugin(idp)
        u = User(first_name='John', last_name='Doe')
        assert idp.status['before_init_document'] == True
        assert idp.status['after_init_document'] == True
                 


from enum import Enum
from functools import wraps
from pkg_resources import iter_entry_points
import yaml


class Hook(Enum):
    BEFORE_CONNECT = 'before_connect'
    AFTER_CONNECT = 'after_connect'
    BEFORE_INIT_DOCUMENT = 'before_init_document'
    AFTER_INIT_DOCUMENT = 'after_init_document'
    BEFORE_SAVE_DOCUMENT = 'before_save_document'
    AFTER_SAVE_DOCUMENT = 'after_save_document'
    BEFORE_DELETE_DOCUMENT = 'before_delete_document'
    AFTER_DELETE_DOCUMENT = 'after_delete_document'
    BEFORE_QUERY_DOCUMENT = 'before_query_document'
    AFTER_QUERY_DOCUMENT = 'after_query_document'
    VALIDATE_DOCUMENT = 'validate_document'
    ON_DOCUMENT_VALIDATION_ERROR = 'on_document_validation_error'
    ON_PLUGIN_LOAD = 'on_plugin_load'
    ON_PLUGIN_UNLOAD = 'on_plugin_unload'
    ON_PLUGIN_ERROR = 'on_plugin_error'
    BEFORE_CLOSE = 'before_close'
    AFTER_CLOSE = 'after_close'

class PluginRegistry:

    def __init__(self):
        self._plugins = []

    def register_plugin(self, plugin):
        try:
            self._plugins.append(plugin)
            if hasattr(plugin, Hook.ON_PLUGIN_LOAD.value):
                plugin.on_plugin_load()
        except Exception as e:
            self.dispatch(Hook.ON_PLUGIN_ERROR, plugin, e)

    def unregister_plugin(self, plugin):
        try:
            self._plugins.remove(plugin)
            if hasattr(plugin, Hook.ON_PLUGIN_UNLOAD.value):
                plugin.on_plugin_unload()
        except Exception as e:
            self.dispatch(Hook.ON_PLUGIN_ERROR, plugin, e)

    def dispatch(self, hook, *args, **kwargs):
        if not isinstance(hook, Hook):
            raise ValueError(f"Invalid hook: {hook}")
        for plugin in self._plugins:
            hook_method = getattr(plugin, hook.value, None)
            if callable(hook_method):
                try:
                    hook_method(*args, **kwargs)
                except Exception as e:
                    if hook == Hook.VALIDATE_DOCUMENT:
                        self.dispatch(Hook.ON_DOCUMENT_VALIDATION_ERROR, plugin, e)
                    else:
                        self.dispatch(Hook.ON_PLUGIN_ERROR, plugin, e)


    def discover(self):
        for entry_point in iter_entry_points('mongeasy.plugins'):
            plugin = entry_point.load()
            self.register(plugin())

    def load_config(self, config_file):
        with open(config_file, 'r') as f:
            config = yaml.safe_load(f)

            self.apply_config(config)

    def apply_config(self, config):
        # TODO: Validate config
        for plugin in self._plugins:
            plugin_name = type(plugin).__name__
            if plugin_name in config:
                for key, value in config[plugin_name].items():
                    if hasattr(plugin, key):
                        setattr(plugin, key, value)
                    else:
                        self.dispatch(Hook.ON_PLUGIN_ERROR, plugin,
                                        Exception(f"Invalid config key: {key} for plugin {plugin_name}"))


def plugin_dispatcher(hooks):
    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            from mongeasy import registry
            # Pre-hooks
            for hook in hooks:
                if hook.get('when') == 'pre' and (hook.get('condition') is None or hook['condition'](func.__name__)):
                    registry.dispatch(hook['hook'],self, *args, **kwargs)

            # Call the actual function
            result = func(self, *args, **kwargs)

            # Post-hooks
            for hook in hooks:
                if hook.get('when') == 'post' and (hook.get('condition') is None or hook['condition'](func.__name__)):
                    if func.__name__ == "__init__":
                        # If the method is __init__, pass self as the data
                        registry.dispatch(hook['hook'],self)
                    else:
                        registry.dispatch(hook['hook'], self, *args, **kwargs) 
            
            return result
        return wrapper
    return decorator




class SampleSavePlugin:
    def before_save_document(self, *args, **kwargs):
        print(f"before_save_plugin: before_save, data: {args}, {kwargs}")
        
    
    def after_save_document(self, data):
        print(f"after_save_plugin: after_save, data: {data}")
        

        









class LazyResultList:
    """
    A lazy list that wraps a MongoDB cursor and caches results as they are fetched.
    """
    def __init__(self, cursor, doc_class=None):
        self.cursor = cursor
        self.doc_class = doc_class
        self._cache = []

    def __getitem__(self, index):
        if index < 0:
            raise NotImplementedError('Negative indices are not supported')
        if isinstance(index, slice):
            return [self._getitem(i) for i in range(*index.indices(len(self)))]
        elif isinstance(index, int):
            return self._getitem(index)
        else:
            raise TypeError("indices must be integers or slices")

    def _getitem(self, index):
        if isinstance(index, int):
            while len(self._cache) <= index:
                try:
                    if self.doc_class:
                        self._cache.append(self.doc_class(next(self.cursor)))
                    else:
                        self._cache.append(next(self.cursor))
                except StopIteration:
                    raise IndexError('LazyList index out of range')
            return self._cache[index]
        else:
            raise TypeError("indices must be integers or slices")
        
    def __repr__(self):
        return "LazyList({!r})".format(self.cursor)
    
    def __iter__(self):
        for i in range(len(self._cache)):
            yield self._cache[i]
        for item in self.cursor:
            self._cache.append(item)
            yield item
            
    def first(self):
        return self[0]
    
    def next(self):
        """
        Fetch the next item from the MongoDB cursor.
        """
        try:
            item = next(self.cursor)
        except StopIteration:
            raise StopIteration("No more items in the list.")
            
        self._cache.append(item)
        return item


#!/usr/bin/env python

"""Tests for `mongeasy` package."""
# Do this before importing mongeasy as it will try to connect to the database
import os
os.environ['MONGOEASY_CONNECTION_STRING'] = 'mongodb://localhost:27017'
os.environ['MONGOEASY_DATABASE_NAME'] = 'mongeasy_pytest'
from pymongo import MongoClient
import pytest
from mongeasy import create_document_class
from mongeasy.exceptions import MongeasyDBDocumentError
from mongeasy.resultlist import ResultList
from mongeasy.lazy_resultlist import LazyResultList


@pytest.fixture(scope='function')
def clean_mongo():
    # Connect to MongoDB
    client = MongoClient('mongodb://localhost:27017')
    db = client['mongeasy_pytest']

    # Clean all collections in the test database
    for collection_name in db.list_collection_names():
        db[collection_name].delete_many({})



def test_document_init_with_dict(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    stored_user = User.find({'first_name': 'John'}).first()
    assert user.first_name == stored_user.first_name
    assert user.last_name == stored_user.last_name
    assert user.age == stored_user.age


def test_document_init_with_kwargs(clean_mongo):
    User = create_document_class('User', 'users')
    user = User(first_name='Pete', last_name='Doe', age=30)
    user.save()
    stored_user = User.find({'first_name': 'Pete'}).first()
    assert user.first_name == stored_user.first_name
    assert user.last_name == stored_user.last_name
    assert user.age == stored_user.age
    assert True


# def test_document_init_with_positional_arguments(clean_mongo):
#     assert True


# def test_document_init_with_invalid_arguments(clean_mongo):
#     assert True


def test_document_has_changed_with_no_changes(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    assert not user.has_changed()


def test_document_has_changed_with_changes(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    user.age += 1
    assert user.has_changed()


def test_document_has_changed_with_new_document(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})

    assert user.has_changed()


def test_document_is_saved_with_no_changes(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    assert user.is_saved()

def test_document_is_saved_with_changes(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    user.age += 1
    assert not user.is_saved()


def test_document_save_new_document(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    stored_user = User.find({'first_name': 'John'}).first()
    assert user.first_name == stored_user.first_name
    assert user.last_name == stored_user.last_name
    assert user.age == stored_user.age


def test_document_save_existing_document_with_no_changes(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    user.save()
    stored_user = User.find({'first_name': 'John'}).first()
    assert user.first_name == stored_user.first_name
    assert user.last_name == stored_user.last_name
    assert user.age == stored_user.age


def test_document_save_existing_document_with_changes(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    user.age += 1
    user.save()
    stored_user = User.find({'first_name': 'John'}).first()
    assert user.first_name == stored_user.first_name
    assert user.last_name == stored_user.last_name
    assert user.age == stored_user.age


def test_document_reload_existing_document(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    user.age += 1
    user.reload()
    assert user.age == 30


def test_document_reload_unsaved_document(clean_mongo):
    from mongeasy.exceptions import MongeasyDBDocumentError
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    with pytest.raises(MongeasyDBDocumentError):
        user.reload()

def test_document_delete_field(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    user.delete_field('age')
    user.save()
    stored_user = User.find({'first_name': 'John'}).first()
    assert 'age' not in stored_user


def test_document_delete_existing_document(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    user.delete()
    stored_user = User.find({'first_name': 'John'}).first()
    assert stored_user is None


def test_document_delete_unsaved_document(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    with pytest.raises(MongeasyDBDocumentError):
        user.delete()
        stored_user = User.find({'first_name': 'John'}).first()
        assert stored_user is None


def test_document_to_json(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    json_user = user.to_json()
    assert isinstance(json_user, str)
    assert 'John' in json_user
    assert 'Doe' in json_user
    assert '30' in json_user
    assert str(user._id) in json_user


def test_document_get_by_id_with_valid_id(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    stored_user = User.get_by_id(user._id)
    assert user.first_name == stored_user.first_name
    assert user.last_name == stored_user.last_name
    assert user.age == stored_user.age


def test_document_get_by_id_with_invalid_id(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    stored_user = User.get_by_id('123')
    assert stored_user is None


def test_document_insert_many(clean_mongo):
    users = [
    {'first_name': 'Alice', 'last_name': 'Smith', 'age': 25},
    {'first_name': 'Bob', 'last_name': 'Jones', 'age': 35},
    {'first_name': 'Charlie', 'last_name': 'Brown', 'age': 45},
    {'first_name': 'Dave', 'last_name': 'Smith', 'age': 55}
]
    User = create_document_class('User', 'users')
    User.insert_many(users)
    stored_users = User.all()
    assert len(stored_users) == 4

def test_document_find(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    stored_user = User.find({'first_name': 'John'}).first()
    assert user.first_name == stored_user.first_name
    assert user.last_name == stored_user.last_name
    assert user.age == stored_user.age


def test_document_find_in(clean_mongo):
    User = create_document_class('User', 'users')
    user1 = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user1.save()
    user2 = User({'first_name': 'Alice', 'last_name': 'Smith', 'age': 25})
    user2.save()
    user3 = User({'first_name': 'Bob', 'last_name': 'Jones', 'age': 35})
    user3.save()
    stored_users = User.find_in('first_name', ['John', 'Bob'])
    assert len(stored_users) == 2
    assert user1.first_name in [user.first_name for user in stored_users]
    assert user2.first_name not in [user.first_name for user in stored_users]
    assert user3.first_name in [user.first_name for user in stored_users]


def test_document_all(clean_mongo):
    User = create_document_class('User', 'users')
    user1 = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user1.save()
    user2 = User({'first_name': 'Alice', 'last_name': 'Smith', 'age': 25})
    user2.save()
    user3 = User({'first_name': 'Bob', 'last_name': 'Jones', 'age': 35})
    user3.save()
    stored_users = User.all()
    assert len(stored_users) == 3
    assert user1.first_name in [user.first_name for user in stored_users]
    assert user2.first_name in [user.first_name for user in stored_users]
    assert user3.first_name in [user.first_name for user in stored_users]


def test_document_delete(clean_mongo):
    User = create_document_class('User', 'users')
    user1 = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user1.save()
    user2 = User({'first_name': 'Alice', 'last_name': 'Smith', 'age': 25})
    user2.save()
    user3 = User({'first_name': 'Bob', 'last_name': 'Jones', 'age': 35})
    user3.save()
    User.delete_many({'first_name': 'Alice'})
    stored_users = User.all()
    assert len(stored_users) == 2
    assert user1.first_name in [user.first_name for user in stored_users]
    assert user2.first_name not in [user.first_name for user in stored_users]
    assert user3.first_name in [user.first_name for user in stored_users]


def test_document_document_count(clean_mongo):
    User = create_document_class('User', 'users')
    user1 = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user1.save()
    user2 = User({'first_name': 'Alice', 'last_name': 'Smith', 'age': 25})
    user2.save()
    user3 = User({'first_name': 'Bob', 'last_name': 'Jones', 'age': 35})
    user3.save()
    count = User.document_count()
    assert count == 3


####################
# Test Query Class #
####################
from mongeasy.core import Query

def test_eq():
    query = Query('name == "John"')
    assert query.to_mongo_query() == {'name': {'$eq': 'John'}}

def test_ne():
    query = Query('name != "John"')
    assert query.to_mongo_query() == {'name': {'$ne': 'John'}}

def test_lt():
    query = Query('age < 30')
    assert query.to_mongo_query() == {'age': {'$lt': 30}}

def test_gt():
    query = Query('age > 30')
    assert query.to_mongo_query() == {'age': {'$gt': 30}}

def test_lte():
    query = Query('age <= 30')
    assert query.to_mongo_query() == {'age': {'$lte': 30}}

def test_gte():
    query = Query('age >= 30')
    assert query.to_mongo_query() == {'age': {'$gte': 30}}

def test_and():
    query = Query('name == "John" and age < 30')
    assert query.to_mongo_query() == {'$and': [{'name': {'$eq': 'John'}}, {'age': {'$lt': 30}}]}

def test_or():
    query = Query('name == "John" or age < 30')
    assert query.to_mongo_query() == {'$or': [{'name': {'$eq': 'John'}}, {'age': {'$lt': 30}}]}

def test_complex():
    query = Query('(name == "John" or age < 30) or (name == "Jane" and age > 20)')
    assert query.to_mongo_query() == {'$or': [{'$or': [{'name': {'$eq': 'John'}}, {'age': {'$lt': 30}}]}, {'$and': [{'name': {'$eq': 'Jane'}}, {'age': {'$gt': 20}}]}]}

def test_subdocument_eq():
    query = Query('friends.age == 32')
    assert query.to_mongo_query() == {'friends.age': {'$eq': 32}}

def test_subdocument_ne():
    query = Query('friends.age != 32')
    assert query.to_mongo_query() == {'friends.age': {'$ne': 32}}

def test_subdocument_and():
    query = Query('age > 25 and friends.age == 32')
    assert query.to_mongo_query() == {'$and': [{'age': {'$gt': 25}}, {'friends.age': {'$eq': 32}}]}

def test_subdocument_or():
    query = Query('age > 25 or friends.age == 32')
    assert query.to_mongo_query() == {'$or': [{'age': {'$gt': 25}}, {'friends.age': {'$eq': 32}}]}

def test_subdocument_complex():
    query = Query('(age > 25 and friends.age == 32) or (name == "Jane" and friends.name == "John")')
    assert query.to_mongo_query() == {'$or': [{'$and': [{'age': {'$gt': 25}}, {'friends.age': {'$eq': 32}}]}, {'$and': [{'name': {'$eq': 'Jane'}}, {'friends.name': {'$eq': 'John'}}]}]}

def test_invalid_query_string():
    with pytest.raises(ValueError) as excinfo:
        q = Query("age >> 25")  # Invalid operator
        q.to_mongo_query()
    assert "Invalid query string" in str(excinfo.value)

def test_invalid_operator():
    with pytest.raises(ValueError) as excinfo:
        q = Query("age <> 25")  # Invalid operator
        q.to_mongo_query()
    assert "Invalid query string" in str(excinfo.value)

def test_invalid_variable():
    with pytest.raises(ValueError) as excinfo:
        q = Query("25 > 25")  # Invalid variable (number instead of a string)
        q.to_mongo_query()
    assert "Invalid query string" in str(excinfo.value)

def test_unmatched_parentheses():
    with pytest.raises(ValueError) as excinfo:
        q = Query("(age > 25 and (name == 'John'")  # Unmatched parentheses
        q.to_mongo_query()
    assert "Invalid query string" in str(excinfo.value)


####################
# Test Lazyness    #
####################

def test_lazy_result_list(clean_mongo):
    client = MongoClient('mongodb://localhost:27017')
    db = client['mongeasy_pytest']
    collection = db['users']
    documents = [{'name': f'user{i}', 'age': i} for i in range(10)]
    collection.insert_many(documents)
    User = create_document_class('User', 'users')
    cursor = collection.find({})
    lazy_result_list = LazyResultList(cursor, User)

    # Test that we can iterate over the LazyResultList
    for i, document in enumerate(lazy_result_list):
        assert document['name'] == f'user{i}'
        assert document['age'] == i

 
def test_lazy_result_list_with_query(clean_mongo):
    client = MongoClient('mongodb://localhost:27017')
    db = client['mongeasy_pytest']
    collection = db['users']
    documents = [{'name': f'user{i}', 'age': i} for i in range(10)]
    collection.insert_many(documents)
    User = create_document_class('User', 'users')
    q = Query('age > 5')
    lazy_result_list = User.find(q, lazy=True)

    # Test that we can iterate over the LazyResultList
    for i, document in enumerate(lazy_result_list):
        assert document['name'] == f'user{i+6}'
        assert document['age'] == i+6


"""Top-level package for Mongeasy."""

__author__ = """Joakim Wassberg"""
__email__ = 'joakim.wassberg@arthead.se'

from mongeasy.connection import MongeasyConnection
from mongeasy.exceptions import MongeasyDBConnectionError
from mongeasy.dynamics import create_document_class
from mongeasy.core import Query
from mongeasy.plugins.registry import PluginRegistry, Hook, plugin_dispatcher

registry = PluginRegistry()
connection = MongeasyConnection(registry=registry)


def connect(connection_str: str=None, db_name: str=None):
    connection.connect(connection_str, db_name)

def disconnect():
    connection.disconnect()
#


import os
import import_string
import yaml
import pymongo
from mongeasy.exceptions import MongeasyDBConnectionError
from mongeasy.plugins.registry import Hook


class MongeasyConnection:
    def __init__(self, uri=None, database=None, connection_options=None, registry=None):
        self.uri = uri
        self.database = database
        self.connection_options = {} if connection_options is None else connection_options
        self.connection_options.setdefault('maxPoolSize', 100)
        self.connection_options.setdefault('retryWrites', True)
        self.connection_options.setdefault('retryReads', True)
        self.check_env()
        self.check_conf_file()

        if self.uri is None:
            raise ValueError("No connection string provided.")
        if self.database is None:
            raise ValueError("No database name provided.")
        
        self.registry = registry

        if registry is not None:
                registry.dispatch(Hook.BEFORE_CONNECT)

        try:
            self.client = pymongo.MongoClient(self.uri, **self.connection_options)
            self.db = self.client[self.database]
        except (pymongo.errors.ConnectionFailure,
                pymongo.errors.ServerSelectionTimeoutError,
                pymongo.errors.ConfigurationError,
                pymongo.errors.InvalidURI) as e:
            raise MongeasyDBConnectionError(f'Failed to connect to database: {e}')
        if registry is not None:
                registry.dispatch(Hook.AFTER_CONNECT)
        
    def check_env(self):
        if self.uri is None:
            self.uri = os.environ.get('MONGOEASY_CONNECTION_STRING')
        if self.database is None:
            self.database = os.environ.get('MONGOEASY_DATABASE_NAME')
        if self.connection_options is None:
            self.connection_options = os.environ.get('MONGOEASY_CONNECTION_OPTIONS')

    def check_conf_file(self):
        config_path = os.path.join(os.getcwd(), 'mongeasy_config.yml')

        # Check if the configuration file exists
        if os.path.exists(config_path):
            # Read the configuration from the file
            with open(config_path, 'r') as f:
                config = yaml.safe_load(f)
                if config is None:
                    config = {}
                db_config = config.get('db_config', {})

            if db_config != {}:
    
                if self.uri is None:
                    self.uri = db_config.get('uri', None)
                if self.database is None:
                    self.database = db_config.get('database', None)
                if self.connection_options is None:
                    self.connection_options = db_config.get('connection_options', None)

            # Handle plugins
            from mongeasy import registry
            plugins_paths = config.get('plugins', [])
            for plugin_path in plugins_paths:
                plugin = import_string(plugin_path)()
                registry.register_plugin(plugin)


    def connect(self, uri=None, database=None, connection_options=None):
        if uri is not None:
            self.uri = uri
        if database is not None:
            self.database = database
        if connection_options is not None:
            self.connection_options = connection_options

        self.client = pymongo.MongoClient(self.uri, **self.connection_options)
        self.db = self.client[self.database]
        return self.db

    def disconnect(self):
        if self.registry is not None:
                self.registry.dispatch(Hook.BEFORE_CLOSE)
        self.client.close()
        if self.registry is not None:
                self.registry.dispatch(Hook.AFTER_CLOSE)



class SampleDeletePlugin:
    def before_delete_document(self, *args, **kwargs):
        print(f"before_delete_document_plugin: before_delete_document")
        
    
    def after_delete_document(self, *args, **kwargs):
        print(f"after_delete_document_plugin: after_delete_document")


class SampleInitPlugin:
    def before_init_document(self, *args, **kwargs):
        print(f"before_init_document_plugin: before_init_document, data: {args}, {kwargs}")
        
    def after_init_document(self, data):
        print(f"after_init_document_plugin: after_init_document, data: {data}")



"""
This module contains the base class making document objects behave like dicts.

MIT License

Copyright (c) 2023 Joakim Wassberg

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation 
files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, 
copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom 
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE 
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from copy import deepcopy


class BaseDict(dict):
    """
    This class is used as a base class for the Document class
    Implements the dict methods
    """
    def __setitem__(self, key, item):
        self.__dict__[key] = item

    def __getitem__(self, key):
        return self.__dict__[key]

    def __repr__(self):
        return repr(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __delitem__(self, key):
        del self.__dict__[key]

    def clear(self):
        return self.__dict__.clear()

    def copy(self):
        # get the class of the current object
        cls = self.__class__
        return cls(self.__dict__.copy())

    def deepcopy(self):
        # get the class of the current object
        cls = self.__class__
        return cls(deepcopy(self.__dict__))

    def has_key(self, k):
        return k in self.__dict__

    def update(self, *args, **kwargs):
        return self.__dict__.update(*args, **kwargs)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def pop(self, *args):
        return self.__dict__.pop(*args)

    def popitem(self):
        return self.__dict__.popitem()
    
    def setdefault(self, *args):
        return self.__dict__.setdefault(*args)
    
    def __cmp__(self, dict_):
        return super().__cmp__(self.__dict__, dict_)

    def __contains__(self, item):
        return item in self.__dict__


class SampleConnectionPlugin:
    def before_connect(self):
        print("before_connect_plugin: before_connect")
    
    def after_connect(self):
        print("after_connect_plugin: after_connect")
    
    def before_close(self):
        print("before_close_plugin: before_close")
    
    def after_close(self):
        print("after_close_plugin: after_close")